@@ -12,7 +12,7 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-import subprocess, os.path, tempfile
+import subprocess, os.path, re
 
 from .. import mlog
 from .. import coredata
@@ -32,6 +32,7 @@ from .compilers import (
     vs32_instruction_set_args,
     vs64_instruction_set_args,
     ArmCompiler,
+    ArmclangCompiler,
     ClangCompiler,
     Compiler,
     CompilerArgs,
@@ -45,6 +46,8 @@ from .compilers import (
 
 class CCompiler(Compiler):
     library_dirs_cache = {}
+    program_dirs_cache = {}
+    find_library_cache = {}
 
     def __init__(self, exelist, version, is_cross, exe_wrapper=None, **kwargs):
         # If a child ObjC or CPP class has already set it, don't set it ourselves
@@ -91,7 +94,7 @@ class CCompiler(Compiler):
         # Almost every compiler uses this for disabling warnings
         return ['-w']
 
-    def get_soname_args(self, prefix, shlib_name, suffix, path, soversion, is_shared_module):
+    def get_soname_args(self, *args):
         return []
 
     def split_shlib_to_parts(self, fname):
@@ -166,11 +169,12 @@ class CCompiler(Compiler):
         env = os.environ.copy()
         env['LC_ALL'] = 'C'
         stdo = Popen_safe(self.exelist + ['--print-search-dirs'], env=env)[1]
+        paths = []
         for line in stdo.split('\n'):
             if line.startswith('libraries:'):
                 libstr = line.split('=', 1)[1]
-                return libstr.split(':')
-        return []
+                paths = [os.path.realpath(p) for p in libstr.split(':')]
+        return paths
 
     def get_library_dirs(self):
         key = tuple(self.exelist)
@@ -178,6 +182,27 @@ class CCompiler(Compiler):
             self.library_dirs_cache[key] = self.get_library_dirs_real()
         return self.library_dirs_cache[key][:]
 
+    def get_program_dirs_real(self):
+        env = os.environ.copy()
+        env['LC_ALL'] = 'C'
+        stdo = Popen_safe(self.exelist + ['--print-search-dirs'], env=env)[1]
+        paths = []
+        for line in stdo.split('\n'):
+            if line.startswith('programs:'):
+                libstr = line.split('=', 1)[1]
+                paths = [os.path.realpath(p) for p in libstr.split(':')]
+        return paths
+
+    def get_program_dirs(self):
+        '''
+        Programs used by the compiler. Also where toolchain DLLs such as
+        libstdc++-6.dll are found with MinGW.
+        '''
+        key = tuple(self.exelist)
+        if key not in self.program_dirs_cache:
+            self.program_dirs_cache[key] = self.get_program_dirs_real()
+        return self.program_dirs_cache[key][:]
+
     def get_pic_args(self):
         return ['-fPIC']
 
@@ -257,7 +282,10 @@ class CCompiler(Compiler):
         else:
             cmdlist = [binary_name]
         mlog.debug('Running test binary command: ' + ' '.join(cmdlist))
-        pe = subprocess.Popen(cmdlist)
+        try:
+            pe = subprocess.Popen(cmdlist)
+        except Exception as e:
+            raise EnvironmentException('Could not invoke sanity test executable: %s.' % str(e))
         pe.wait()
         if pe.returncode != 0:
             raise EnvironmentException('Executables created by {0} compiler {1} are not runnable.'.format(self.language, self.name_string()))
@@ -266,6 +294,12 @@ class CCompiler(Compiler):
         code = 'int main(int argc, char **argv) { int class=0; return class; }\n'
         return self.sanity_check_impl(work_dir, environment, 'sanitycheckc.c', code)
 
+    def check_header(self, hname, prefix, env, extra_args=None, dependencies=None):
+        fargs = {'prefix': prefix, 'header': hname}
+        code = '''{prefix}
+        #include <{header}>'''
+        return self.compiles(code.format(**fargs), env, extra_args, dependencies)
+
     def has_header(self, hname, prefix, env, extra_args=None, dependencies=None):
         fargs = {'prefix': prefix, 'header': hname}
         code = '''{prefix}
@@ -322,13 +356,13 @@ class CCompiler(Compiler):
         if not self.is_cross:
             if mode == 'preprocess':
                 # Add CPPFLAGS from the env.
-                args += env.coredata.external_preprocess_args[self.language]
+                args += env.coredata.get_external_preprocess_args(self.language)
             elif mode == 'compile':
                 # Add CFLAGS/CXXFLAGS/OBJCFLAGS/OBJCXXFLAGS from the env
-                args += env.coredata.external_args[self.language]
+                args += env.coredata.get_external_args(self.language)
             elif mode == 'link':
                 # Add LDFLAGS from the env
-                args += env.coredata.external_link_args[self.language]
+                args += env.coredata.get_external_link_args(self.language)
         args += self.get_compiler_check_args()
         # extra_args must override all other arguments, so we add them last
         args += extra_args
@@ -537,8 +571,9 @@ class CCompiler(Compiler):
             if p.returncode != 0:
                 raise EnvironmentException('Could not get define {!r}'.format(dname))
         # Get the preprocessed value after the delimiter,
-        # minus the extra newline at the end
-        return p.stdo.split(delim + '\n')[-1][:-1]
+        # minus the extra newline at the end and
+        # merge string literals.
+        return CCompiler.concatenate_string_literals(p.stdo.split(delim + '\n')[-1][:-1])
 
     def get_return_value(self, fname, rtype, prefix, env, extra_args, dependencies):
         if rtype == 'string':
@@ -808,17 +843,11 @@ class CCompiler(Compiler):
             raise AssertionError('BUG: unknown libtype {!r}'.format(libtype))
         return prefixes, suffixes
 
-    def find_library(self, libname, env, extra_dirs, libtype='default'):
-        # These libraries are either built-in or invalid
-        if libname in self.ignore_libs:
-            return []
+    def find_library_real(self, libname, env, extra_dirs, code, libtype):
         # First try if we can just add the library as -l.
-        code = 'int main(int argc, char **argv) { return 0; }'
-        if extra_dirs and isinstance(extra_dirs, str):
-            extra_dirs = [extra_dirs]
         # Gcc + co seem to prefer builtin lib dirs to -L dirs.
         # Only try to find std libs if no extra dirs specified.
-        if not extra_dirs and libtype == 'default':
+        if not extra_dirs:
             args = ['-l' + libname]
             if self.links(code, env, extra_args=args):
                 return args
@@ -834,10 +863,41 @@ class CCompiler(Compiler):
             for suffix in suffixes:
                 for prefix in prefixes:
                     trial = os.path.join(d, prefix + libname + '.' + suffix)
-                    if os.path.isfile(trial):
+                    # as well as checking the path, we need to check compilation
+                    # with link-whole, as static libs (.a) need to be checked
+                    # to ensure they are the right architecture, e.g. 32bit or
+                    # 64-bit. Just a normal test link won't work as the .a file
+                    # doesn't seem to be checked by linker if there are no
+                    # unresolved symbols from the main C file.
+                    extra_link_args = self.get_link_whole_for([trial])
+                    extra_link_args = self.linker_to_compiler_args(extra_link_args)
+                    if (os.path.isfile(trial) and
+                            self.links(code, env,
+                                       extra_args=extra_link_args)):
                         return [trial]
+        # XXX: For OpenBSD and macOS we (may) need to search for libfoo.x.y.z.dylib
         return None
 
+    def find_library_impl(self, libname, env, extra_dirs, code, libtype):
+        # These libraries are either built-in or invalid
+        if libname in self.ignore_libs:
+            return []
+        if isinstance(extra_dirs, str):
+            extra_dirs = [extra_dirs]
+        key = (tuple(self.exelist), libname, tuple(extra_dirs), code, libtype)
+        if key not in self.find_library_cache:
+            value = self.find_library_real(libname, env, extra_dirs, code, libtype)
+            self.find_library_cache[key] = value
+        else:
+            value = self.find_library_cache[key]
+        if value is None:
+            return None
+        return value[:]
+
+    def find_library(self, libname, env, extra_dirs, libtype='default'):
+        code = 'int main(int argc, char **argv) { return 0; }'
+        return self.find_library_impl(libname, env, extra_dirs, code, libtype)
+
     def thread_flags(self, env):
         if for_haiku(self.is_cross, env):
             return []
@@ -889,6 +949,16 @@ class CCompiler(Compiler):
         code = 'int main(int argc, char **argv) { return 0; }'
         return self.has_arguments(args, env, code, mode='link')
 
+    @staticmethod
+    def concatenate_string_literals(s):
+        pattern = re.compile(r'(?P<pre>.*([^\\]")|^")(?P<str1>([^\\"]|\\.)*)"\s+"(?P<str2>([^\\"]|\\.)*)(?P<post>".*)')
+        ret = s
+        m = pattern.match(ret)
+        while m:
+            ret = ''.join(m.group('pre', 'str1', 'str2', 'post'))
+            m = pattern.match(ret)
+        return ret
+
 class ClangCCompiler(ClangCompiler, CCompiler):
     def __init__(self, exelist, version, clang_type, is_cross, exe_wrapper=None, **kwargs):
         CCompiler.__init__(self, exelist, version, is_cross, exe_wrapper, **kwargs)
@@ -899,10 +969,12 @@ class ClangCCompiler(ClangCompiler, CCompiler):
                           '3': default_warn_args + ['-Wextra', '-Wpedantic']}
 
     def get_options(self):
-        return {'c_std': coredata.UserComboOption('c_std', 'C language standard to use',
-                                                  ['none', 'c89', 'c99', 'c11',
-                                                   'gnu89', 'gnu99', 'gnu11'],
-                                                  'none')}
+        opts = CCompiler.get_options(self)
+        opts.update({'c_std': coredata.UserComboOption('c_std', 'C language standard to use',
+                                                       ['none', 'c89', 'c99', 'c11',
+                                                        'gnu89', 'gnu99', 'gnu11'],
+                                                       'none')})
+        return opts
 
     def get_option_compile_args(self, options):
         args = []
@@ -921,6 +993,34 @@ class ClangCCompiler(ClangCompiler, CCompiler):
         return basic
 
 
+class ArmclangCCompiler(ArmclangCompiler, CCompiler):
+    def __init__(self, exelist, version, is_cross, exe_wrapper=None, **kwargs):
+        CCompiler.__init__(self, exelist, version, is_cross, exe_wrapper, **kwargs)
+        ArmclangCompiler.__init__(self)
+        default_warn_args = ['-Wall', '-Winvalid-pch']
+        self.warn_args = {'1': default_warn_args,
+                          '2': default_warn_args + ['-Wextra'],
+                          '3': default_warn_args + ['-Wextra', '-Wpedantic']}
+
+    def get_options(self):
+        opts = CCompiler.get_options(self)
+        opts.update({'c_std': coredata.UserComboOption('c_std', 'C language standard to use',
+                                                       ['none', 'c90', 'c99', 'c11',
+                                                        'gnu90', 'gnu99', 'gnu11'],
+                                                       'none')})
+        return opts
+
+    def get_option_compile_args(self, options):
+        args = []
+        std = options['c_std']
+        if std.value != 'none':
+            args.append('-std=' + std.value)
+        return args
+
+    def get_option_link_args(self, options):
+        return []
+
+
 class GnuCCompiler(GnuCompiler, CCompiler):
     def __init__(self, exelist, version, gcc_type, is_cross, exe_wrapper=None, defines=None, **kwargs):
         CCompiler.__init__(self, exelist, version, is_cross, exe_wrapper, **kwargs)
@@ -931,10 +1031,11 @@ class GnuCCompiler(GnuCompiler, CCompiler):
                           '3': default_warn_args + ['-Wextra', '-Wpedantic']}
 
     def get_options(self):
-        opts = {'c_std': coredata.UserComboOption('c_std', 'C language standard to use',
-                                                  ['none', 'c89', 'c99', 'c11',
-                                                   'gnu89', 'gnu99', 'gnu11'],
-                                                  'none')}
+        opts = CCompiler.get_options(self)
+        opts.update({'c_std': coredata.UserComboOption('c_std', 'C language standard to use',
+                                                       ['none', 'c89', 'c99', 'c11',
+                                                        'gnu89', 'gnu99', 'gnu11'],
+                                                       'none')})
         if self.gcc_type == GCC_MINGW:
             opts.update({
                 'c_winlibs': coredata.UserArrayOption('c_winlibs', 'Standard Win libraries to link against',
@@ -967,11 +1068,12 @@ class ElbrusCCompiler(GnuCCompiler, ElbrusCompiler):
 
     # It does support some various ISO standards and c/gnu 90, 9x, 1x in addition to those which GNU CC supports.
     def get_options(self):
-        opts = {'c_std': coredata.UserComboOption('c_std', 'C language standard to use',
-                                                  ['none', 'c89', 'c90', 'c9x', 'c99', 'c1x', 'c11',
-                                                   'gnu89', 'gnu90', 'gnu9x', 'gnu99', 'gnu1x', 'gnu11',
-                                                   'iso9899:2011', 'iso9899:1990', 'iso9899:199409', 'iso9899:1999'],
-                                                  'none')}
+        opts = CCompiler.get_options(self)
+        opts.update({'c_std': coredata.UserComboOption('c_std', 'C language standard to use',
+                                                       ['none', 'c89', 'c90', 'c9x', 'c99', 'c1x', 'c11',
+                                                        'gnu89', 'gnu90', 'gnu9x', 'gnu99', 'gnu1x', 'gnu11',
+                                                        'iso9899:2011', 'iso9899:1990', 'iso9899:199409', 'iso9899:1999'],
+                                                       'none')})
         return opts
 
     # Elbrus C compiler does not have lchmod, but there is only linker warning, not compiler error.
@@ -994,13 +1096,14 @@ class IntelCCompiler(IntelCompiler, CCompiler):
                           '3': default_warn_args + ['-Wextra', '-Wpedantic']}
 
     def get_options(self):
+        opts = CCompiler.get_options(self)
         c_stds = ['c89', 'c99']
         g_stds = ['gnu89', 'gnu99']
         if version_compare(self.version, '>=16.0.0'):
             c_stds += ['c11']
-        opts = {'c_std': coredata.UserComboOption('c_std', 'C language standard to use',
-                                                  ['none'] + c_stds + g_stds,
-                                                  'none')}
+        opts.update({'c_std': coredata.UserComboOption('c_std', 'C language standard to use',
+                                                       ['none'] + c_stds + g_stds,
+                                                       'none')})
         return opts
 
     def get_option_compile_args(self, options):
@@ -1139,10 +1242,11 @@ class VisualStudioCCompiler(CCompiler):
         return []
 
     def get_options(self):
-        return {'c_winlibs': coredata.UserArrayOption('c_winlibs',
-                                                      'Windows libs to link against.',
-                                                      msvc_winlibs)
-                }
+        opts = CCompiler.get_options(self)
+        opts.update({'c_winlibs': coredata.UserArrayOption('c_winlibs',
+                                                           'Windows libs to link against.',
+                                                           msvc_winlibs)})
+        return opts
 
     def get_option_link_args(self, options):
         return options['c_winlibs'].value[:]
@@ -1263,9 +1367,10 @@ class ArmCCompiler(ArmCompiler, CCompiler):
         ArmCompiler.__init__(self)
 
     def get_options(self):
-        opts = {'c_std': coredata.UserComboOption('c_std', 'C language standard to use',
-                                                  ['none', 'c90', 'c99'],
-                                                  'none')}
+        opts = CCompiler.get_options(self)
+        opts.update({'c_std': coredata.UserComboOption('c_std', 'C language standard to use',
+                                                       ['none', 'c90', 'c99'],
+                                                       'none')})
         return opts
 
     def get_option_compile_args(self, options):
