@@ -13,7 +13,7 @@
 # limitations under the License.
 
 from . import mlog
-import pickle, os, uuid
+import pickle, os, uuid, shlex
 import sys
 from pathlib import PurePath
 from collections import OrderedDict
@@ -22,7 +22,7 @@ from .mesonlib import default_libdir, default_libexecdir, default_prefix
 import ast
 import argparse
 
-version = '0.46.1'
+version = '0.47.0'
 backendlist = ['ninja', 'vs', 'vs2010', 'vs2015', 'vs2017', 'xcode']
 
 default_yielding = False
@@ -105,6 +105,22 @@ class UserIntegerOption(UserOption):
         except ValueError:
             raise MesonException('Value string "%s" is not convertable to an integer.' % valuestring)
 
+class UserUmaskOption(UserIntegerOption):
+    def __init__(self, name, description, value, yielding=None):
+        super().__init__(name, description, 0, 0o777, value, yielding)
+        self.choices = ['preserve', '0000-0777']
+
+    def validate_value(self, value):
+        if value is None or value == 'preserve':
+            return None
+        return super().validate_value(value)
+
+    def toint(self, valuestring):
+        try:
+            return int(valuestring, 8)
+        except ValueError as e:
+            raise MesonException('Invalid mode: {}'.format(e))
+
 class UserComboOption(UserOption):
     def __init__(self, name, description, choices, value, yielding=None):
         super().__init__(name, description, choices, yielding)
@@ -122,33 +138,38 @@ class UserComboOption(UserOption):
         return value
 
 class UserArrayOption(UserOption):
-    def __init__(self, name, description, value, **kwargs):
+    def __init__(self, name, description, value, shlex_split=False, user_input=False, **kwargs):
         super().__init__(name, description, kwargs.get('choices', []), yielding=kwargs.get('yielding', None))
-        self.value = self.validate_value(value, user_input=False)
+        self.shlex_split = shlex_split
+        self.value = self.validate_value(value, user_input=user_input)
 
     def validate_value(self, value, user_input=True):
         # User input is for options defined on the command line (via -D
         # options). Users can put their input in as a comma separated
         # string, but for defining options in meson_options.txt the format
         # should match that of a combo
-        if not user_input:
-            if isinstance(value, str):
-                if not value.startswith('['):
-                    raise MesonException('Valuestring does not define an array: ' + value)
-                newvalue = ast.literal_eval(value)
-            else:
-                newvalue = value
-        else:
-            assert isinstance(value, str)
+        if not user_input and isinstance(value, str) and not value.startswith('['):
+            raise MesonException('Value does not define an array: ' + value)
+
+        if isinstance(value, str):
             if value.startswith('['):
                 newvalue = ast.literal_eval(value)
+            elif value == '':
+                newvalue = []
             else:
-                newvalue = [v.strip() for v in value.split(',')]
-                if len(set(newvalue)) != len(newvalue):
-                    mlog.log(mlog.red('DEPRECATION:'), '''Duplicated values in an array type is deprecated.
-This will become a hard error in the future.''')
-        if not isinstance(newvalue, list):
+                if self.shlex_split:
+                    newvalue = shlex.split(value)
+                else:
+                    newvalue = [v.strip() for v in value.split(',')]
+        elif isinstance(value, list):
+            newvalue = value
+        else:
             raise MesonException('"{0}" should be a string array, but it is not'.format(str(newvalue)))
+
+        if len(set(newvalue)) != len(newvalue):
+            msg = 'Duplicated values in array option "%s" is deprecated. ' \
+                  'This will become a hard error in the future.' % (self.name)
+            mlog.log(mlog.red('DEPRECATION:'), msg)
         for i in newvalue:
             if not isinstance(i, str):
                 raise MesonException('String array element "{0}" is not a string.'.format(str(newvalue)))
@@ -159,6 +180,21 @@ This will become a hard error in the future.''')
                     ', '.join(bad), ', '.join(self.choices)))
         return newvalue
 
+class UserFeatureOption(UserComboOption):
+    static_choices = ['enabled', 'disabled', 'auto']
+
+    def __init__(self, name, description, value, yielding=None):
+        super().__init__(name, description, self.static_choices, value, yielding)
+
+    def is_enabled(self):
+        return self.value == 'enabled'
+
+    def is_disabled(self):
+        return self.value == 'disabled'
+
+    def is_auto(self):
+        return self.value == 'auto'
+
 # This class contains all data that must persist over multiple
 # invocations of Meson. It is roughly the same thing as
 # cmakecache.
@@ -166,21 +202,23 @@ This will become a hard error in the future.''')
 class CoreData:
 
     def __init__(self, options):
-        self.guid = str(uuid.uuid4()).upper()
+        self.lang_guids = {
+            'default': '8BC9CEB8-8B4A-11D0-8D11-00A0C91BC942',
+            'c': '8BC9CEB8-8B4A-11D0-8D11-00A0C91BC942',
+            'cpp': '8BC9CEB8-8B4A-11D0-8D11-00A0C91BC942',
+            'test': '3AC096D0-A1C2-E12C-1390-A8335801FDAB',
+            'directory': '2150E333-8FDC-42A3-9474-1A3956D46DE8',
+        }
         self.test_guid = str(uuid.uuid4()).upper()
         self.regen_guid = str(uuid.uuid4()).upper()
         self.target_guids = {}
         self.version = version
-        self.init_builtins(options)
-        self.init_backend_options(self.builtins['backend'].value)
+        self.init_builtins()
+        self.backend_options = {}
         self.user_options = {}
         self.compiler_options = {}
         self.base_options = {}
-        # These external_*args, are set via env vars CFLAGS, LDFLAGS, etc
-        # but only when not cross-compiling.
         self.external_preprocess_args = {} # CPPFLAGS only
-        self.external_args = {} # CPPFLAGS + CFLAGS
-        self.external_link_args = {} # CFLAGS + LDFLAGS (with MSVC: only LDFLAGS)
         self.cross_file = self.__load_cross_file(options.cross_file)
         self.wrap_mode = options.wrap_mode
         self.compilers = OrderedDict()
@@ -215,7 +253,7 @@ class CoreData:
         if os.path.isabs(filename):
             return filename
         path_to_try = os.path.abspath(filename)
-        if os.path.exists(path_to_try):
+        if os.path.isfile(path_to_try):
             return path_to_try
         if sys.platform != 'win32':
             paths = [
@@ -223,7 +261,7 @@ class CoreData:
             ] + os.environ.get('XDG_DATA_DIRS', '/usr/local/share:/usr/share').split(':')
             for path in paths:
                 path_to_try = os.path.join(path, 'meson', 'cross', filename)
-                if os.path.exists(path_to_try):
+                if os.path.isfile(path_to_try):
                     return path_to_try
             raise MesonException('Cannot find specified cross file: ' + filename)
 
@@ -271,27 +309,30 @@ class CoreData:
             value = value[skip:]
         return value
 
-    def init_builtins(self, options):
+    def init_builtins(self):
+        # Create builtin options with default values
         self.builtins = {}
-        # Sanitize prefix
-        options.prefix = self.sanitize_prefix(options.prefix)
-        # Initialize other builtin options
+        prefix = get_builtin_option_default('prefix')
         for key in get_builtin_options():
-            if hasattr(options, key):
-                value = getattr(options, key)
-                value = self.sanitize_dir_option_value(options.prefix, key, value)
-                setattr(options, key, value)
-            else:
-                value = get_builtin_option_default(key, prefix=options.prefix)
+            value = get_builtin_option_default(key, prefix)
             args = [key] + builtin_options[key][1:-1] + [value]
             self.builtins[key] = builtin_options[key][0](*args)
 
     def init_backend_options(self, backend_name):
-        self.backend_options = {}
         if backend_name == 'ninja':
-            self.backend_options['backend_max_links'] = UserIntegerOption('backend_max_links',
-                                                                          'Maximum number of linker processes to run or 0 for no limit',
-                                                                          0, None, 0)
+            self.backend_options['backend_max_links'] = \
+                UserIntegerOption(
+                    'backend_max_links',
+                    'Maximum number of linker processes to run or 0 for no '
+                    'limit',
+                    0, None, 0)
+        elif backend_name.startswith('vs'):
+            self.backend_options['backend_startup_project'] = \
+                UserStringOption(
+                    'backend_startup_project',
+                    'Default project to execute in Visual Studio',
+                    '')
+
 
     def get_builtin_option(self, optname):
         if optname in self.builtins:
@@ -315,6 +356,61 @@ class CoreData:
                 return opt.validate_value(override_value)
         raise MesonException('Tried to validate unknown option %s.' % option_name)
 
+    def get_external_args(self, lang):
+        return self.compiler_options[lang + '_args'].value
+
+    def get_external_link_args(self, lang):
+        return self.compiler_options[lang + '_link_args'].value
+
+    def get_external_preprocess_args(self, lang):
+        return self.external_preprocess_args[lang]
+
+    def merge_user_options(self, options):
+        for (name, value) in options.items():
+            if name not in self.user_options:
+                self.user_options[name] = value
+            else:
+                oldval = self.user_options[name]
+                if type(oldval) != type(value):
+                    self.user_options[name] = value
+
+    def set_options(self, options, subproject=''):
+        # Set prefix first because it's needed to sanitize other options
+        prefix = self.builtins['prefix'].value
+        if 'prefix' in options:
+            prefix = self.sanitize_prefix(options['prefix'])
+            self.builtins['prefix'].set_value(prefix)
+            for key in builtin_dir_noprefix_options:
+                if key not in options:
+                    self.builtins[key].set_value(get_builtin_option_default(key, prefix))
+
+        unknown_options = []
+        for k, v in options.items():
+            if k == 'prefix':
+                pass
+            elif k in self.builtins:
+                tgt = self.builtins[k]
+                tgt.set_value(self.sanitize_dir_option_value(prefix, k, v))
+            elif k in self.backend_options:
+                tgt = self.backend_options[k]
+                tgt.set_value(v)
+            elif k in self.user_options:
+                tgt = self.user_options[k]
+                tgt.set_value(v)
+            elif k in self.compiler_options:
+                tgt = self.compiler_options[k]
+                tgt.set_value(v)
+            elif k in self.base_options:
+                tgt = self.base_options[k]
+                tgt.set_value(v)
+            else:
+                unknown_options.append(k)
+
+        if unknown_options:
+            unknown_options = ', '.join(sorted(unknown_options))
+            sub = 'In subproject {}: '.format(subproject) if subproject else ''
+            mlog.warning('{}Unknown options: "{}"'.format(sub, unknown_options))
+
 def load(build_dir):
     filename = os.path.join(build_dir, 'meson-private', 'coredata.dat')
     load_fail_msg = 'Coredata file {!r} is corrupted. Try with a fresh build tree.'.format(filename)
@@ -332,10 +428,19 @@ def load(build_dir):
 
 def save(obj, build_dir):
     filename = os.path.join(build_dir, 'meson-private', 'coredata.dat')
+    prev_filename = filename + '.prev'
+    tempfilename = filename + '~'
     if obj.version != version:
         raise MesonException('Fatal version mismatch corruption.')
-    with open(filename, 'wb') as f:
+    if os.path.exists(filename):
+        import shutil
+        shutil.copyfile(filename, prev_filename)
+    with open(tempfilename, 'wb') as f:
         pickle.dump(obj, f)
+        f.flush()
+        os.fsync(f.fileno())
+    os.replace(tempfilename, filename)
+    return filename
 
 def get_builtin_options():
     return list(builtin_options.keys())
@@ -345,12 +450,14 @@ def is_builtin_option(optname):
 
 def get_builtin_option_choices(optname):
     if is_builtin_option(optname):
-        if builtin_options[optname][0] == UserStringOption:
-            return None
+        if builtin_options[optname][0] == UserComboOption:
+            return builtin_options[optname][2]
         elif builtin_options[optname][0] == UserBooleanOption:
             return [True, False]
+        elif builtin_options[optname][0] == UserFeatureOption:
+            return UserFeatureOption.static_choices
         else:
-            return builtin_options[optname][2]
+            return None
     else:
         raise RuntimeError('Tried to get the supported values for an unknown builtin option \'%s\'.' % optname)
 
@@ -368,19 +475,17 @@ def get_builtin_option_action(optname):
         return 'store_true'
     return None
 
-def get_builtin_option_default(optname, prefix='', noneIfSuppress=False):
+def get_builtin_option_default(optname, prefix=''):
     if is_builtin_option(optname):
         o = builtin_options[optname]
         if o[0] == UserComboOption:
             return o[3]
-        if optname in builtin_dir_noprefix_options:
-            if noneIfSuppress:
-                # Return None if argparse defaulting should be suppressed for
-                # this option (so we can determine the default later based on
-                # prefix)
-                return None
-            elif prefix in builtin_dir_noprefix_options[optname]:
-                return builtin_dir_noprefix_options[optname][prefix]
+        if o[0] == UserIntegerOption:
+            return o[4]
+        try:
+            return builtin_dir_noprefix_options[optname][prefix]
+        except KeyError:
+            pass
         return o[2]
     else:
         raise RuntimeError('Tried to get the default value for an unknown builtin option \'%s\'.' % optname)
@@ -402,11 +507,7 @@ def add_builtin_argument(p, name):
         kwargs['action'] = b
     if c and not b:
         kwargs['choices'] = c
-    default = get_builtin_option_default(name, noneIfSuppress=True)
-    if default is not None:
-        kwargs['default'] = default
-    else:
-        kwargs['default'] = argparse.SUPPRESS
+    kwargs['default'] = argparse.SUPPRESS
     kwargs['dest'] = name
 
     cmdline_name = get_builtin_option_cmdline_name(name)
@@ -418,32 +519,30 @@ def register_builtin_arguments(parser):
     parser.add_argument('-D', action='append', dest='projectoptions', default=[], metavar="option",
                         help='Set the value of an option, can be used several times to set multiple options.')
 
-def filter_builtin_options(args, original_args):
-    """Filter out any builtin arguments passed as -- instead of -D.
+def create_options_dict(options):
+    result = {}
+    for o in options:
+        try:
+            (key, value) = o.split('=', 1)
+        except ValueError:
+            raise MesonException('Option {!r} must have a value separated by equals sign.'.format(o))
+        result[key] = value
+    return result
 
-    Error if an argument is passed with -- and -D
-    """
+def parse_cmd_line_options(args):
+    args.cmd_line_options = create_options_dict(args.projectoptions)
+
+    # Merge builtin options set with --option into the dict.
     for name in builtin_options:
-        # Check if user passed --option. Cannot use hasattr(args, name) here
-        # because they are all set with default value if user didn't pass it.
-        cmdline_name = get_builtin_option_cmdline_name(name)
-        has_dashdash = any([a.startswith(cmdline_name) for a in original_args])
-
-        # Chekc if user passed -Doption=value
-        has_dashd = any([a.startswith('{}='.format(name)) for a in args.projectoptions])
-
-        # Passing both is ambigous, abort
-        if has_dashdash and has_dashd:
-            raise MesonException(
-                'Got argument {0} as both -D{0} and {1}. Pick one.'.format(name, cmdline_name))
-
-        # Pretend --option never existed
-        if has_dashdash:
-            args.projectoptions.append('{}={}'.format(name, getattr(args, name)))
-        if hasattr(args, name):
+        value = getattr(args, name, None)
+        if value is not None:
+            if name in args.cmd_line_options:
+                cmdline_name = get_builtin_option_cmdline_name(name)
+                raise MesonException(
+                    'Got argument {0} as both -D{0} and {1}. Pick one.'.format(name, cmdline_name))
+            args.cmd_line_options[name] = value
             delattr(args, name)
 
-
 builtin_options = {
     'buildtype':  [UserComboOption, 'Build type to use.', ['plain', 'debug', 'debugoptimized', 'release', 'minsize'], 'debug'],
     'strip':      [UserBooleanOption, 'Strip targets on install.', False],
@@ -468,6 +567,8 @@ builtin_options = {
     'backend':         [UserComboOption, 'Backend to use.', backendlist, 'ninja'],
     'stdsplit':        [UserBooleanOption, 'Split stdout and stderr in test logs.', True],
     'errorlogs':       [UserBooleanOption, "Whether to print the logs from failing tests.", True],
+    'install_umask':   [UserUmaskOption, 'Default umask to apply on permissions of installed files.', '022'],
+    'auto_features':   [UserFeatureOption, "Override value of all 'auto' features.", 'auto'],
 }
 
 # Special prefix-dependent defaults for installation directories that reside in
